/**************************************************************************
*
*	Main.cpp
*
*	Main source file for cooperative robots sample programme
*
*	Author: Saxon Druce
*
*	Copyright (c) 2005 Block Software Pty Ltd
*
**************************************************************************/

/*

We could write a programme with a single main thread and have it make two
connections to the server, and therefore control two robots, each taking
turns. However, this would put the robots at a disadvantage, because while 
one robot is being controlled the other isn't. Since a limited number of
messages can be sent to the server per second, each robot's turn could
take a significant amount of time. So what we do instead is spawn two
threads and have each thread control a single robot.

The cooperative robots sample consists of two robots - a leader and a 
follower. The leader works similarly to a simple robot, ie it wanders
around randomly until it finds a robot, and then tries to attack it.
The follower wanders around randomly until the leader finds a robot,
and then tries to attack the same robot as the leader is attacking.

*/

/*************************************************************************/
// Include files
/*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <time.h>
#include <windows.h>
#include <process.h>

#include "Controller.h"
#include "CriticalSection.h"

/*************************************************************************/
// Defines
/*************************************************************************/

/*************************************************************************/
// Local types
/*************************************************************************/

// The class which controls the robots
class Robots_t
{
public:

	// Constructor.
	Robots_t(const std::string& ServerName)
	{
		m_ServerName = ServerName;

		m_FollowerRobotID = -1;

		m_LeaderEnemyPositionSet = false;
	}

	// Run() - Runs the robots.
	void Run()
	{
		// create a thread for each of the robots
		unsigned int ThreadID;
		HANDLE LeaderThreadHandle = (HANDLE)_beginthreadex(
			NULL, 0, LeaderStaticFunc, this, 0, &ThreadID);
		HANDLE FollowerThreadHandle = (HANDLE)_beginthreadex(
			NULL, 0, FollowerStaticFunc, this, 0, &ThreadID);

		// wait until they finish
		WaitForSingleObject(LeaderThreadHandle, INFINITE);
		WaitForSingleObject(FollowerThreadHandle, INFINITE);
	}

private:

	// LeaderStaticFunc() - Static method which executes in the thread
	// which controls the leader robot, passing the call on to the class
	// method.
	static unsigned int __stdcall LeaderStaticFunc(void *Param)
	{
		((Robots_t*)Param)->Leader();
		return 0;
	}

	// FollowerStaticFunc() - Static method which executes in the thread
	// which controls the follower robot, passing the call on to the class
	// method.
	static unsigned int __stdcall FollowerStaticFunc(void *Param)
	{
		((Robots_t*)Param)->Follower();
		return 0;
	}

	// Leader() - The method which executes in the thread which controls
	// the leader robot.
	void Leader()
	{
		try
		{
			printf("Leader thread started\n\n");

			// seed random number generator in this thread
			srand((unsigned int)time(NULL) ^ (unsigned int)GetCurrentThreadId());

			// create a controller
			Controller_t Controller;

			// connect to the server
			Controller.Connect(m_ServerName, 9386);

			// get the playing field size
			double PlayingFieldWidth, PlayingFieldHeight;
			Controller.GetPlayingFieldSize(
				PlayingFieldWidth, PlayingFieldHeight);

			// create a robot
			int RobotID = Controller.CreateRobot();
			printf("Leader Robot ID is %d\n", RobotID);

			// set name and owner
			Controller.SetName("Leader Robot");
			Controller.SetOwner("Saxon");

			// set colour
			Controller.SetColour(0, 100, 100);

			// get initial amount of money
			printf("Leader Robot has $%d to spend on equipment\n", 
				Controller.GetMoney());

			// buy equipment
			Controller.PurchaseEngine(3, 75);
			Controller.PurchaseSensor(10, 60, 60);
			Controller.PurchaseBattery(50);
			Controller.PurchaseTurret(60, 75);
			Controller.PurchaseWeapon(1.3, 4, 2, 4.5);
			Controller.PurchaseSolarPanel(1.5);

			// get amount of money left
			printf("Leader Robot has $%d remaining after purchasing equipment\n", 
				Controller.GetMoney());

			// enter the simulation
			Controller.EnterSimulation();

			// set up a value to hold the position which the robot is
			// trying to get to, and initialise it with the robot's
			// current position
			Vec_t TargetPosition = Controller.GetPosition();

			// set up a value to hold the time at which we selected the target
			time_t TargetSelectionTime = time(NULL);

			while (true)
			{
				// get the robot's current position
				Vec_t CurrentPosition = Controller.GetPosition();

				// get the robot's current orientation
				double CurrentOrientation = Controller.GetOrientation();

				// sense other robots
				std::vector<Controller_t::SensedRobot_t> Robots = 
					Controller.GetSensedRobots();

				if (!Robots.empty() && Robots[0].m_RobotID == GetFollowerRobotID())
				{
					// the first robot is the follower robot, so remove it,
					// since we usually attack the first robot, and we don't
					// want to attack our own follower!
					Robots.erase(Robots.begin());
				}

				if (Robots.empty())
				{
					// we couldn't see any robots, so simply wander randomly
					// around the playing field, in the hope that we'll
					// eventually find another robot

					// indicate that we're not targetting a robot
					ClearLeaderEnemyPosition();

					// work out the vector from the current position to the
					// target position
					Vec_t TargetVector = TargetPosition - CurrentPosition;

					if (TargetVector.Length() < 1.0 ||
						time(NULL) - TargetSelectionTime > 10)
					{
						// we're within 1 metre of the target so assume we've
						// reached it, or it's been 10 seconds since we chose 
						// the target so assume we're in deadlock - therefore 
						// pick a new random target position, which is at least 
						// 2 metres from the edge of the playing field
						TargetPosition.x = RandomBetween(
							-(PlayingFieldWidth / 2.0 - 2.0), 
							(PlayingFieldWidth / 2.0 - 2.0));
						TargetPosition.y = RandomBetween(
							-(PlayingFieldHeight / 2.0 - 2.0), 
							(PlayingFieldHeight / 2.0 - 2.0));

						printf(
							"New leader target position is %.1f, %.1f\n",
							TargetPosition.x,
							TargetPosition.y);
						
						// recalculate the target vector since it is used below
						TargetVector = TargetPosition - CurrentPosition;

						// update target selection time
						TargetSelectionTime = time(NULL);
					}

					// work out the desired orientation to point towards the
					// target
					double DesiredOrientation = TargetVector.GetAngle();

					// set it
					Controller.SetDesiredOrientation(
						DesiredOrientation);

					// work out the difference between the current orientation and
					// the desired orientation
					double OrientationDifference = 
						DesiredOrientation - CurrentOrientation;
					
					// wrap to -pi to pi range
					if (OrientationDifference < -M_PI)
						OrientationDifference += 2 * M_PI;
					else if (OrientationDifference >= M_PI)
						OrientationDifference -= 2 * M_PI;

					if (fabs(OrientationDifference) < 5 * M_PI / 180)
					{
						// we're within 5 degrees of the desired orientation, so
						// set the throttle to full forwards
						Controller.SetThrottle(100.0);
					}
					else
					{
						// set the throttle to stay still while we turn
						Controller.SetThrottle(0.0);
					}

					// set sensor orientation to 90 degrees further clockwise
					// than it currently is, which should result in a constant
					// clockwise spin as fast as possible (assuming we aren't
					// able to spin by 90 degrees between iterations of this loop)
					Controller.SetDesiredSensorOrientation(
						Controller.GetSensorOrientation() + M_PI / 2);
				}
				else
				{
					// we can see at least one robot, so attack the first one

					// set the position of the target so that the follower 
					// targets it too
					SetLeaderEnemyPosition(Robots[0].m_Position);

					// work out the vector from the current position to the
					// enemy robot position
					Vec_t EnemyVector = Robots[0].m_Position - CurrentPosition;

					// work out the desired orientation to point towards the
					// enemy robot
					double DesiredOrientation = EnemyVector.GetAngle();

					// set it
					Controller.SetDesiredOrientation(
						DesiredOrientation);

					// work out the difference between the current orientation and
					// the desired orientation
					double OrientationDifference = 
						DesiredOrientation - CurrentOrientation;
					
					// wrap to -pi to pi range
					if (OrientationDifference < -M_PI)
						OrientationDifference += 2 * M_PI;
					else if (OrientationDifference >= M_PI)
						OrientationDifference -= 2 * M_PI;

					if (fabs(OrientationDifference) < 5 * M_PI / 180)
					{
						// we're within 5 degrees of the desired orientation

						// work out how far we are from the enemy
						double EnemyDistance = EnemyVector.Length();

						if (EnemyDistance < 5)
						{
							// within 5 metres of the enemy, so move
							// backwards to keep away from it (we'll
							// be too easy to hit if we get too close)
							Controller.SetThrottle(-100.0);
						}
						else if (EnemyDistance < 6)
						{
							// between 5 and 6 metres from the enemy,
							// so stay still
							Controller.SetThrottle(0.0);
						}
						else
						{
							// more than 6 metres from the enemy, so
							// move towards it
							Controller.SetThrottle(100.0);
						}

						if (EnemyDistance < 6)
						{
							// we're within a reasonable range of the enemy,
							// so fire the weapon if it's fully charged
							if (Controller.GetWeaponCharge() >= 100)
								Controller.FireWeapon();
						}
					}
					else
					{
						// set the throttle to stay still while we turn
						Controller.SetThrottle(0.0);
					}

					// set the sensor angle required to keep it pointing 
					// at the enemy, to make sure it doesn't get away
					Controller.SetDesiredSensorOrientation(
						DesiredOrientation - CurrentOrientation);
				}
			}
		}
		catch (std::exception& e)
		{
			printf("\nAn exception occurred with the leader: %s\n\n", e.what());

			// if the leader has died, don't let the follower keep shooting at 
			// a robot which has probably moved on
			ClearLeaderEnemyPosition();
		}
	}

	// Leader() - The method which executes in the thread which controls
	// the leader robot.
	void Follower()
	{
		try
		{
			printf("Follower thread started\n\n");

			// seed random number generator in this thread
			srand((unsigned int)time(NULL) ^ (unsigned int)GetCurrentThreadId());

			// create a controller
			Controller_t Controller;

			// connect to the server
			Controller.Connect(m_ServerName, 9386);

			// get the playing field size
			double PlayingFieldWidth, PlayingFieldHeight;
			Controller.GetPlayingFieldSize(
				PlayingFieldWidth, PlayingFieldHeight);

			// create a robot
			int RobotID = Controller.CreateRobot();
			printf("Follower Robot ID is %d\n", RobotID);

			// set the follower robot ID - this is so that the leader
			// doesn't try to attack us
			SetFollowerRobotID(RobotID);

			// set name and owner
			Controller.SetName("Follower Robot");
			Controller.SetOwner("Saxon");

			// set colour
			Controller.SetColour(0, 100, 100);

			// get initial amount of money
			printf("Follower Robot has $%d to spend on equipment\n", 
				Controller.GetMoney());

			// buy equipment
			Controller.PurchaseEngine(3, 75);
			Controller.PurchaseSensor(10, 60, 60);
			Controller.PurchaseBattery(50);
			Controller.PurchaseTurret(60, 75);
			Controller.PurchaseWeapon(1.3, 4, 2, 4.5);
			Controller.PurchaseSolarPanel(1.5);

			// get amount of money left
			printf("Follower Robot has $%d remaining after purchasing equipment\n", 
				Controller.GetMoney());

			// enter the simulation
			Controller.EnterSimulation();

			// set up a value to hold the position which the robot is
			// trying to get to, and initialise it with the robot's
			// current position
			Vec_t TargetPosition = Controller.GetPosition();

			// set up a value to hold the time at which we selected the target
			time_t TargetSelectionTime = time(NULL);

			while (true)
			{
				// get the robot's current position
				Vec_t CurrentPosition = Controller.GetPosition();

				// get the robot's current orientation
				double CurrentOrientation = Controller.GetOrientation();

				// see if the leader is currently targetting a robot
				Vec_t LeaderEnemyPosition;
				bool LeaderEnemyPositionSet = 
					GetLeaderEnemyPosition(LeaderEnemyPosition);

				if (!LeaderEnemyPositionSet)
				{
					// leader isn't targetting any robots, so simply wander 
					// randomly around the playing field - the point of this
					// is so that we're not a sitting duck while we wait for
					// the leader to find something, but it can result in a
					// deadlock if both leader and follower choose random
					// points which make them run into each other.

					// work out the vector from the current position to the
					// target position
					Vec_t TargetVector = TargetPosition - CurrentPosition;

					if (TargetVector.Length() < 1.0 ||
						time(NULL) - TargetSelectionTime > 10)
					{
						// we're within 1 metre of the target so assume we've
						// reached it, or it's been 10 seconds since we chose 
						// the target so assume we're in deadlock - therefore 
						// pick a new random target position, which is at least 
						// 2 metres from the edge of the playing field
						TargetPosition.x = RandomBetween(
							-(PlayingFieldWidth / 2.0 - 2.0), 
							(PlayingFieldWidth / 2.0 - 2.0));
						TargetPosition.y = RandomBetween(
							-(PlayingFieldHeight / 2.0 - 2.0), 
							(PlayingFieldHeight / 2.0 - 2.0));

						printf(
							"New follower target position is %.1f, %.1f\n",
							TargetPosition.x,
							TargetPosition.y);
						
						// recalculate the target vector since it is used below
						TargetVector = TargetPosition - CurrentPosition;

						// update target selection time
						TargetSelectionTime = time(NULL);
					}

					// work out the desired orientation to point towards the
					// target
					double DesiredOrientation = TargetVector.GetAngle();

					// set it
					Controller.SetDesiredOrientation(
						DesiredOrientation);

					// work out the difference between the current orientation and
					// the desired orientation
					double OrientationDifference = 
						DesiredOrientation - CurrentOrientation;
					
					// wrap to -pi to pi range
					if (OrientationDifference < -M_PI)
						OrientationDifference += 2 * M_PI;
					else if (OrientationDifference >= M_PI)
						OrientationDifference -= 2 * M_PI;

					if (fabs(OrientationDifference) < 5 * M_PI / 180)
					{
						// we're within 5 degrees of the desired orientation, so
						// set the throttle to full forwards
						Controller.SetThrottle(100.0);
					}
					else
					{
						// set the throttle to stay still while we turn
						Controller.SetThrottle(0.0);
					}
				}
				else
				{
					// the leader is targetting a robot, so attack it too

					// work out the vector from the current position to the
					// enemy robot position
					Vec_t EnemyVector = LeaderEnemyPosition - CurrentPosition;

					// work out the desired orientation to point towards the
					// enemy robot
					double DesiredOrientation = EnemyVector.GetAngle();

					// set it
					Controller.SetDesiredOrientation(
						DesiredOrientation);

					// work out the difference between the current orientation and
					// the desired orientation
					double OrientationDifference = 
						DesiredOrientation - CurrentOrientation;
					
					// wrap to -pi to pi range
					if (OrientationDifference < -M_PI)
						OrientationDifference += 2 * M_PI;
					else if (OrientationDifference >= M_PI)
						OrientationDifference -= 2 * M_PI;

					if (fabs(OrientationDifference) < 5 * M_PI / 180)
					{
						// we're within 5 degrees of the desired orientation

						// work out how far we are from the enemy
						double EnemyDistance = EnemyVector.Length();

						if (EnemyDistance < 5)
						{
							// within 5 metres of the enemy, so move
							// backwards to keep away from it (we'll
							// be too easy to hit if we get too close)
							Controller.SetThrottle(-100.0);
						}
						else if (EnemyDistance < 6)
						{
							// between 5 and 6 metres from the enemy,
							// so stay still
							Controller.SetThrottle(0.0);
						}
						else
						{
							// more than 6 metres from the enemy, so
							// move towards it
							Controller.SetThrottle(100.0);
						}

						if (EnemyDistance < 6)
						{
							// we're within a reasonable range of the enemy,
							// so fire the weapon if it's fully charged
							if (Controller.GetWeaponCharge() >= 100)
								Controller.FireWeapon();
						}
					}
					else
					{
						// set the throttle to stay still while we turn
						Controller.SetThrottle(0.0);
					}
				}
			}
		}
		catch (std::exception& e)
		{
			printf("\nAn exception occurred with the follower: %s\n\n", e.what());
		}
	}

	// GetFollowerRobotID() - Returns the follower robot's ID, in a 
	// thread-safe way.
	int GetFollowerRobotID()
	{
		Lock_t Lock(m_CriticalSection);
		return m_FollowerRobotID;
	}

	// SetFollowerRobotID() - Sets the follower robot's ID, in a 
	// thread-safe way.
	void SetFollowerRobotID(int FollowerRobotID)
	{
		Lock_t Lock(m_CriticalSection);
		m_FollowerRobotID = FollowerRobotID;
	}

	// SetLeaderEnemyPosition() - Sets the position of the robot which the
	// leader is targetting, in a thread-safe way.
	void SetLeaderEnemyPosition(const Vec_t& LeaderEnemyPosition)
	{
		Lock_t Lock(m_CriticalSection);
		m_LeaderEnemyPosition = LeaderEnemyPosition;
		m_LeaderEnemyPositionSet = true;
	}

	// ClearLeaderEnemyPosition() - Indicates that the leader robot isn't 
	// currently targetting another robot, in a thread-safe way.
	void ClearLeaderEnemyPosition()
	{
		Lock_t Lock(m_CriticalSection);
		m_LeaderEnemyPositionSet = false;
	}

	// GetLeaderEnemyPosition() - Returns true if the leader is currently
	// targetting another robot, false if not. If true is returned, the
	// enemy's position is returned via LeaderEnemyPosition. Performs the
	// operation in a thread-safe way.
	bool GetLeaderEnemyPosition(Vec_t& LeaderEnemyPosition)
	{
		Lock_t Lock(m_CriticalSection);
		LeaderEnemyPosition = m_LeaderEnemyPosition;
		return m_LeaderEnemyPositionSet;
	}

	// RandomBetween() - Helper function to return a random value between the 
	// two bounds.
	static double RandomBetween(double LowerBound, double UpperBound)
	{
		return LowerBound + rand() * (UpperBound - LowerBound) / RAND_MAX;
	}

	// the name (or IP) of the server to connect to
	std::string m_ServerName;

	// a critical section used to lock variables which both threads acccess
	CriticalSection_t m_CriticalSection;

	// the follower robot's ID, or -1 if not set yet
	int m_FollowerRobotID;

	// the position of the robot which the leader robots is currently
	// attacking
	Vec_t m_LeaderEnemyPosition;

	// true if the leader robot's enemy's position has been set
	bool m_LeaderEnemyPositionSet;
};

/*************************************************************************/
// Local functions
/*************************************************************************/

/*************************************************************************/
// Functions
/*************************************************************************/

/**************************************************************************
*
*	function:	main()
*
*	desc:		Programme entry point
*
*	notes:		
*
**************************************************************************/

void main(int NumArgs, char** Args)
{
	try
	{
		printf("Cooperative Robots Sample\n\n");

		// setup default values of parameters
		std::string ServerName = "localhost";

		// expect arguments to be in the following format:
		// /a a_value /b b_value
		// so there should be an even number of them, after the first one
		// which is the exe filename
		if ((NumArgs - 1) % 2 != 0)
			throw std::runtime_error("Expected an even number of arguments");
		
		// now loop over them
		int i;
		for (i = 1; i < NumArgs; i += 2)
		{
			if (!strcmp(Args[i], "/s"))
			{
				// /s = server address (a string)
				ServerName = Args[i + 1];
			}
			else
			{
				throw std::runtime_error(
					std::string() + "Invalid argument " + Args[i]);
			}
		}

		printf("Use ctrl+C to stop\n\n");

		// run the robots
		Robots_t Robots(ServerName);
		Robots.Run();

		printf("\nDone\n\n");
	}
	catch (std::exception& e)
	{
		printf("\nAn exception occurred: %s\n\n", e.what());
	}
}

/*************************************************************************/
