/**************************************************************************
*
*	SimpleRobot.cs
*
*	Main source file for simple robot sample programme (in C#)
*
*	Author: Saxon Druce
*
*	Copyright (c) 2004 Block Software Pty Ltd
*
**************************************************************************/

/*

We could write a programme with a single main thread and have it make two
connections to the server, and therefore control two robots, each taking
turns. However, this would put the robots at a disadvantage, because while 
one robot is being controlled the other isn't. Since a limited number of
messages can be sent to the server per second, each robot's turn could
take a significant amount of time. So what we do instead is spawn two
threads and have each thread control a single robot.

The cooperative robots sample consists of two robots - a leader and a 
follower. The leader works similarly to a simple robot, ie it wanders
around randomly until it finds a robot, and then tries to attack it.
The follower wanders around randomly until the leader finds a robot,
and then tries to attack the same robot as the leader is attacking.

*/

using System;
using System.Threading;
using System.Collections;
using ControllerNet;
using VecNet;

namespace CooperativeRobotsNet
{
	/// <summary>
	/// The main class.
	/// </summary>
	class CooperativeRobots
	{
		/// <summary>
		/// The class which controls the robots
		/// </summary>
		class Robots
		{
			/// <summary>
			/// Constructor
			/// </summary>
			public Robots(string serverName)
			{
				m_serverName = serverName;

				m_followerRobotID = -1;

				m_leaderEnemyPositionSet = false;
			}

			/// <summary>
			/// Runs the robots.
			/// </summary>
			public void Run()
			{
				// create a thread for each of the robots
				Thread leaderThread = new Thread(new ThreadStart(Leader));
				Thread followerThread = new Thread(new ThreadStart(Follower));

				// start them both
				// Note: I was seeing occasional failures (about 1/3 of the 
				// time) with a file not found exception in whichever thread 
				// started second, when accessing ControllerNet.dll. Obviously 
				// the file is there otherwise the first thread wouldn't have
				// been able to load it. The sleep of 100 milliseconds seems to
				// fix it - I put it down to a case of '.NET weirdness'. Maybe
				// .NET locks access to an assembly while loading it?
				leaderThread.Start();
				Thread.Sleep(100);
				followerThread.Start();

				// wait for them both to finish
				leaderThread.Join();
				followerThread.Join();
			}

			/// <summary>
			/// The method which executes in the thread which controls the 
			/// leader robot.
			/// </summary>
			private void Leader()
			{
				try
				{
					Console.WriteLine("Leader thread started\n");

					// create a controller
					Controller controller = new Controller();

					// connect
					controller.Connect(m_serverName, 9386);

					// get the playing field size
					double playingFieldWidth, playingFieldHeight;
					controller.GetPlayingFieldSize(
						out playingFieldWidth, out playingFieldHeight);

					// create a robot
					int robotID = controller.CreateRobot();
					Console.WriteLine("Leader Robot ID is {0}", robotID);

					// set name and owner
					controller.SetName("Leader Robot C#");
					controller.SetOwner("Saxon");

					// set colour
					controller.SetColour(0, 100, 100);

					// get initial amount of money
					Console.WriteLine(
						"Leader Robot has ${0} to spend on equipment", 
						controller.GetMoney());

					// buy equipment
					controller.PurchaseEngine(3, 75);
					controller.PurchaseSensor(10, 60, 60);
					controller.PurchaseBattery(50);
					controller.PurchaseTurret(0, 0);
					controller.PurchaseWeapon(1.3, 4, 2, 4.5);
					controller.PurchaseSolarPanel(1.5);

					// get amount of money left
					Console.WriteLine(
						"Leader Robot has ${0} remaining after purchasing equipment", 
						controller.GetMoney());

					// enter the simulation
					controller.EnterSimulation();

					// set up a value to hold the position which the robot is
					// trying to get to, and initialise it with the robot's
					// current position
					Vec targetPosition = controller.GetPosition();

					// set up a value to hold the time at which we selected the target
					DateTime TargetSelectionTime = DateTime.Now;

					while (true)
					{
						// get the robot's current position
						Vec currentPosition = controller.GetPosition();

						// get the robot's current orientation
						double currentOrientation = controller.GetOrientation();

						// sense other robots
						ArrayList robots = controller.GetSensedRobots();

						if (robots.Count > 0 && 
							(robots[0] as SensedRobot).m_RobotID == GetFollowerRobotID())
						{
							// the first robot is the follower robot, so remove it,
							// since we usually attack the first robot, and we don't
							// want to attack our own follower!
							robots.RemoveAt(0);
						}

						if (robots.Count == 0)
						{
							// we couldn't see any robots, so simply wander randomly
							// around the playing field, in the hope that we'll
							// eventually find another robot

							// indicate that we're not targetting a robot
							ClearLeaderEnemyPosition();

							// work out the vector from the current position to the
							// target position
							Vec targetVector = targetPosition - currentPosition;

							if (targetVector.Length() < 1.0 ||
								(DateTime.Now - TargetSelectionTime).TotalSeconds > 10)
							{
								// we're within 1 metre of the target so assume we've
								// reached it, or it's been 10 seconds since we chose 
								// the target so assume we're in deadlock - therefore 
								// pick a new random target position, which is at least 
								// 2 metres from the edge of the playing field
								targetPosition.x = RandomBetween(
									-(playingFieldWidth / 2.0 - 2.0), 
									(playingFieldWidth / 2.0 - 2.0));
								targetPosition.y = RandomBetween(
									-(playingFieldHeight / 2.0 - 2.0), 
									(playingFieldHeight / 2.0 - 2.0));

								Console.WriteLine(
									"New leader target position is {0:f1}, {0:f1}",
									targetPosition.x,
									targetPosition.y);
						
								// recalculate the target vector since it is used below
								targetVector = targetPosition - currentPosition;

								// update target selection time
								TargetSelectionTime = DateTime.Now;
							}

							// work out the desired orientation to point towards the
							// target
							double desiredOrientation = targetVector.GetAngle();

							// set it
							controller.SetDesiredOrientation(
								desiredOrientation);

							// work out the difference between the current orientation and
							// the desired orientation
							double orientationDifference = 
								desiredOrientation - currentOrientation;
					
							// wrap to -pi to pi range
							if (orientationDifference < -Math.PI)
								orientationDifference += 2 * Math.PI;
							else if (orientationDifference >= Math.PI)
								orientationDifference -= 2 * Math.PI;

							if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
							{
								// we're within 5 degrees of the desired orientation, so
								// set the throttle to full forwards
								controller.SetThrottle(100.0);
							}
							else
							{
								// set the throttle to stay still while we turn
								controller.SetThrottle(0.0);
							}

							// set sensor orientation to 90 degrees further clockwise
							// than it currently is, which should result in a constant
							// clockwise spin as fast as possible (assuming we aren't
							// able to spin by 90 degrees between iterations of this loop)
							controller.SetDesiredSensorOrientation(
								controller.GetSensorOrientation() + Math.PI / 2);
						}
						else
						{
							// we can see at least one robot, so attack the first one

							// get it
							SensedRobot enemy = (robots[0] as SensedRobot);

							// set the position of the target so that the follower 
							// targets it too
							SetLeaderEnemyPosition(enemy.m_Position);

							// work out the vector from the current position to the
							// enemy robot position
							Vec enemyVector = 
								enemy.m_Position - currentPosition;

							// work out the desired orientation to point towards the
							// enemy robot
							double desiredOrientation = enemyVector.GetAngle();

							// set it
							controller.SetDesiredOrientation(
								desiredOrientation);

							// work out the difference between the current orientation and
							// the desired orientation
							double orientationDifference = 
								desiredOrientation - currentOrientation;
					
							// wrap to -pi to pi range
							if (orientationDifference < -Math.PI)
								orientationDifference += 2 * Math.PI;
							else if (orientationDifference >= Math.PI)
								orientationDifference -= 2 * Math.PI;

							if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
							{
								// we're within 5 degrees of the desired orientation

								// work out how far we are from the enemy
								double enemyDistance = enemyVector.Length();

								if (enemyDistance < 5)
								{
									// within 5 metres of the enemy, so move
									// backwards to keep away from it (we'll
									// be too easy to hit if we get too close)
									controller.SetThrottle(-100.0);
								}
								else if (enemyDistance < 6)
								{
									// between 5 and 6 metres from the enemy,
									// so stay still
									controller.SetThrottle(0.0);
								}
								else
								{
									// more than 6 metres from the enemy, so
									// move towards it
									controller.SetThrottle(100.0);
								}

								if (enemyDistance < 6)
								{
									// we're within a reasonable range of the enemy,
									// so fire the weapon if it's fully charged
									if (controller.GetWeaponCharge() >= 100)
										controller.FireWeapon();
								}
							}
							else
							{
								// set the throttle to stay still while we turn
								controller.SetThrottle(0.0);
							}

							// set the sensor angle required to keep it pointing 
							// at the enemy, to make sure it doesn't get away
							controller.SetDesiredSensorOrientation(
								desiredOrientation - currentOrientation);
						}
					}
				}
				catch (Exception e)
				{
					Console.WriteLine(
						"An exception occurred with the leader: {0}",
						e.Message);

					// if the leader has died, don't let the follower keep 
					// shooting at a robot which has probably moved on
					ClearLeaderEnemyPosition();
				}
			}

			/// <summary>
			/// The method which executes in the thread which controls the 
			/// follower robot.
			/// </summary>
			private void Follower()
			{
				try
				{
					Console.WriteLine("Follower thread started\n");

					// create a controller
					Controller controller = new Controller();

					// connect
					controller.Connect(m_serverName, 9386);

					// get the playing field size
					double playingFieldWidth, playingFieldHeight;
					controller.GetPlayingFieldSize(
						out playingFieldWidth, out playingFieldHeight);

					// create a robot
					int robotID = controller.CreateRobot();
					Console.WriteLine("Follower Robot ID is {0}", robotID);

					// set the follower robot ID - this is so that the leader
					// doesn't try to attack us
					SetFollowerRobotID(robotID);

					// set name and owner
					controller.SetName("Follower Robot C#");
					controller.SetOwner("Saxon");

					// set colour
					controller.SetColour(0, 100, 100);

					// get initial amount of money
					Console.WriteLine(
						"Follower Robot has ${0} to spend on equipment", 
						controller.GetMoney());

					// buy equipment
					controller.PurchaseEngine(3, 75);
					controller.PurchaseSensor(10, 60, 60);
					controller.PurchaseBattery(50);
					controller.PurchaseTurret(0, 0);
					controller.PurchaseWeapon(1.3, 4, 2, 4.5);
					controller.PurchaseSolarPanel(1.5);

					// get amount of money left
					Console.WriteLine(
						"Follower Robot has ${0} remaining after purchasing equipment", 
						controller.GetMoney());

					// enter the simulation
					controller.EnterSimulation();

					// set up a value to hold the position which the robot is
					// trying to get to, and initialise it with the robot's
					// current position
					Vec targetPosition = controller.GetPosition();

					// set up a value to hold the time at which we selected the target
					DateTime TargetSelectionTime = DateTime.Now;

					while (true)
					{
						// get the robot's current position
						Vec currentPosition = controller.GetPosition();

						// get the robot's current orientation
						double currentOrientation = controller.GetOrientation();

						// see if the leader is currently targetting a robot
						Vec leaderEnemyPosition;
						bool leaderEnemyPositionSet = 
							GetLeaderEnemyPosition(out leaderEnemyPosition);

						if (!leaderEnemyPositionSet)
						{
							// leader isn't targetting any robots, so simply wander 
							// randomly around the playing field - the point of this
							// is so that we're not a sitting duck while we wait for
							// the leader to find something, but it can result in a
							// deadlock if both leader and follower choose random
							// points which make them run into each other.

							// work out the vector from the current position to the
							// target position
							Vec targetVector = targetPosition - currentPosition;

							if (targetVector.Length() < 1.0 ||
								(DateTime.Now - TargetSelectionTime).TotalSeconds > 10)
							{
								// we're within 1 metre of the target so assume we've
								// reached it, or it's been 10 seconds since we chose 
								// the target so assume we're in deadlock - therefore 
								// pick a new random target position, which is at least 
								// 2 metres from the edge of the playing field
								targetPosition.x = RandomBetween(
									-(playingFieldWidth / 2.0 - 2.0), 
									(playingFieldWidth / 2.0 - 2.0));
								targetPosition.y = RandomBetween(
									-(playingFieldHeight / 2.0 - 2.0), 
									(playingFieldHeight / 2.0 - 2.0));

								Console.WriteLine(
									"New leader target position is {0:f1}, {0:f1}",
									targetPosition.x,
									targetPosition.y);
						
								// recalculate the target vector since it is used below
								targetVector = targetPosition - currentPosition;

								// update target selection time
								TargetSelectionTime = DateTime.Now;
							}

							// work out the desired orientation to point towards the
							// target
							double desiredOrientation = targetVector.GetAngle();

							// set it
							controller.SetDesiredOrientation(
								desiredOrientation);

							// work out the difference between the current orientation and
							// the desired orientation
							double orientationDifference = 
								desiredOrientation - currentOrientation;
					
							// wrap to -pi to pi range
							if (orientationDifference < -Math.PI)
								orientationDifference += 2 * Math.PI;
							else if (orientationDifference >= Math.PI)
								orientationDifference -= 2 * Math.PI;

							if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
							{
								// we're within 5 degrees of the desired orientation, so
								// set the throttle to full forwards
								controller.SetThrottle(100.0);
							}
							else
							{
								// set the throttle to stay still while we turn
								controller.SetThrottle(0.0);
							}
						}
						else
						{
							// the leader is targetting a robot, so attack it too

							// work out the vector from the current position to the
							// enemy robot position
							Vec enemyVector = 
								leaderEnemyPosition - currentPosition;

							// work out the desired orientation to point towards the
							// enemy robot
							double desiredOrientation = enemyVector.GetAngle();

							// set it
							controller.SetDesiredOrientation(
								desiredOrientation);

							// work out the difference between the current orientation and
							// the desired orientation
							double orientationDifference = 
								desiredOrientation - currentOrientation;
					
							// wrap to -pi to pi range
							if (orientationDifference < -Math.PI)
								orientationDifference += 2 * Math.PI;
							else if (orientationDifference >= Math.PI)
								orientationDifference -= 2 * Math.PI;

							if (Math.Abs(orientationDifference) < 5 * Math.PI / 180)
							{
								// we're within 5 degrees of the desired orientation

								// work out how far we are from the enemy
								double enemyDistance = enemyVector.Length();

								if (enemyDistance < 5)
								{
									// within 5 metres of the enemy, so move
									// backwards to keep away from it (we'll
									// be too easy to hit if we get too close)
									controller.SetThrottle(-100.0);
								}
								else if (enemyDistance < 6)
								{
									// between 5 and 6 metres from the enemy,
									// so stay still
									controller.SetThrottle(0.0);
								}
								else
								{
									// more than 6 metres from the enemy, so
									// move towards it
									controller.SetThrottle(100.0);
								}

								if (enemyDistance < 6)
								{
									// we're within a reasonable range of the enemy,
									// so fire the weapon if it's fully charged
									if (controller.GetWeaponCharge() >= 100)
										controller.FireWeapon();
								}
							}
							else
							{
								// set the throttle to stay still while we turn
								controller.SetThrottle(0.0);
							}
						}
					}
				}
				catch (Exception e)
				{
					Console.WriteLine(
						"An exception occurred with the follower: {0}",
						e.Message);
				}
			}

			// GetFollowerRobotID() - Returns the follower robot's ID, in a 
			// thread-safe way.
			private int GetFollowerRobotID()
			{
				lock (this)
				{
					return m_followerRobotID;
				}
			}

			// SetFollowerRobotID() - Sets the follower robot's ID, in a 
			// thread-safe way.
			private void SetFollowerRobotID(int followerRobotID)
			{
				lock (this)
				{
					m_followerRobotID = followerRobotID;
				}
			}

			/// <summary>
			/// Sets the position of the robot which the leader is targetting, 
			/// in a thread-safe way.
			/// </summary>
			private void SetLeaderEnemyPosition(Vec leaderEnemyPosition)
			{
				lock (this)
				{
					m_leaderEnemyPosition = leaderEnemyPosition;
					m_leaderEnemyPositionSet = true;
				}
			}

			/// <summary>
			/// Indicates that the leader robot isn't currently targetting 
			/// another robot, in a thread-safe way.
			/// </summary>
			private void ClearLeaderEnemyPosition()
			{
				lock (this)
				{
					m_leaderEnemyPositionSet = false;
				}
			}

			/// <summary>
			/// Returns true if the leader is currently targetting another 
			/// robot, false if not. If true is returned, the enemy's position 
			/// is returned via LeaderEnemyPosition. Performs the operation 
			/// in a thread-safe way.
			/// </summary>
			private bool GetLeaderEnemyPosition(out Vec leaderEnemyPosition)
			{
				lock (this)
				{
					leaderEnemyPosition = m_leaderEnemyPosition;
					return m_leaderEnemyPositionSet;
				}
			}

			/// <summary>
			/// Helper function to return a random value between the two bounds.
			/// </summary>
			double RandomBetween(double lowerBound, double upperBound)
			{
				// since m_random is shared by both robot threads, we need to 
				// lock access to it
				lock (m_random)
				{
					return lowerBound + 
						m_random.NextDouble() * (upperBound - lowerBound);
				}
			}

			/// <summary>
			/// the name (or IP) of the server to connect to
			/// </summary>
			private string m_serverName;

			/// <summary>
			/// random number generator
			/// </summary>
			private Random m_random = new Random();

			/// <summary>
			/// the follower robot's ID, or -1 if not set yet
			/// </summary>
			private int m_followerRobotID;

			/// <summary>
			/// the position of the robot which the leader robots is currently
			/// attacking
			/// </summary>
			private Vec m_leaderEnemyPosition;

			/// <summary>
			/// true if the leader robot's enemy's position has been set
			/// </summary>
			private bool m_leaderEnemyPositionSet;
		}

		/// <summary>
		/// The main entry point for the application.
		/// </summary>
		[STAThread]
		static void Main(string[] args)
		{
			try
			{
				Console.WriteLine("Cooperative Robots Sample\n");

				// setup default values of parameters
				string serverName = "localhost";

				// expect arguments to be in the following format:
				// /a a_value /b b_value
				// so there should be an even number of them
				if (args.Length % 2 != 0)
					throw new Exception("Expected an even number of arguments");

				// now loop over them
				int i;
				for (i = 0; i < args.Length; i += 2)
				{
					if (args[i] == "/s")
					{
						// /s = server address (a string)
						serverName = args[i + 1];
					}
					else
					{
						throw new Exception("Invalid argument " + args[i]);
					}
				}

				Console.WriteLine("Use ctrl+C to stop\n");

				// run the robots
				Robots robots = new Robots(serverName);
				robots.Run();

				Console.WriteLine("\nDone\n");
			}
			catch (Exception e)
			{
				Console.WriteLine(
					"An exception occurred: {0}",
					e.Message);
			}
		}
	}
}
